#!/usr/bin/perl
#
# Tool for applying SpeedStep patches to AICPUPM
# Patching method by flAked: http://www.insanelymac.com/forum/index.php?showtopic=258611
# Credits: flAked
#
# 18/02/2012 - dmazar - v1.0 initial version
#

# credits
print "********************\n";
print "Tool for applying SpeedStep patches to AICPUPM\n";
print "Patching method by flAked: http://www.insanelymac.com/forum/index.php?showtopic=258611\nCredits: flAked\n";
print "********************\n";

# usage print
sub usage {
    print "Usage:\n";
    print "sudo perl patch_aicpupm.pl <AICPUPM_executable> <spec_file>\n";
    print "- replaces wrmsr calls with nop in <AICPUPM_executable>\n";
    print "- positions of wrmsr calls should be specified in <spec_file>\n";
    print "- use correct <spec_file> for specified <AICPUPM_executable> version\n";
    print "\n<spec_file>:\n";
    print "- text file with wrmsr positions (in hex) listed, one per line\n";
    print "- specific to particular <AICPUPM_executable> version\n\n";
    exit;
}

# reads positions from spec file and returns them as list
sub read_spec {
    my $spec_file = shift;
    
    # open spec file or die/quit
    open(my $SPEC, $spec_file) || die "ERROR: Can not open spec. file '$spec_file': $!\n";
    
    print "Reading wrmsr positions from spec file '$spec_file' … ";

    my @positions = (), $lineNo;
    # read line from spec file, skip comments (lines starting with #), put positions in the array
    while(<$SPEC>) {
        chomp;
        # trim whitespaces from beginning
        s/^\s+//;
        # skip comments and empty lines
        next if (m/^\#/ || $_ eq "");
        # convert position from hex 
        my $position = hex;
        if (!$position) {
            print "\n";
            die "ERROR: invalid entry '$_' in spec. file '$spec_file' at line $.";
        }
        push @positions, $position;
    }
    close($SPEC);
    print "OK, found " . @positions . " positions\n";
    return @positions;
}


#
# main prog starts here
#

# check args
if ($#ARGV != 1) {
    usage();
}

my ($AICPUPM_file, $spec_file) = @ARGV;

print "Going to patch '$AICPUPM_file' at positions specified in '$spec_file'\n";

my @positions = read_spec($spec_file);
my $WRMSR1 = 0x0f, $WRMSR2 = 0x30, $NOP = 0x90, $num = 1;

# open AICPUPM file for read and write or die/quit
open(my $AICPUPM, "+< $AICPUPM_file") || die "ERROR: Can not open AICPUPM file '$AICPUPM_file': $!\n";

# go patch ...
my $bytes, $b1, $b2, $posHex;
foreach my $position (@positions) {
    $posHex = sprintf("%#x", $position);
    print("$num) $posHex: ");
    
    # move to position in file and read two bytes
    seek($AICPUPM, $position, SEEK_SET) || die "ERROR: position $posHex can not be reached: $!\n";
    read($AICPUPM, $bytes, 2);
    
    # we have 2 bytes - unpack them
    my ($b1, $b2) = unpack('C2', $bytes);
    
    # make some checking
    if ($b1 == $NOP && $b2 == $NOP) {
        # nops already there - patched
        print "already patched\n";
    } elsif ($b1 != $WRMSR1 || $b2 != $WRMSR2) {
        # no wrmsr - not the file we can work with
        die "ERROR: wrmsr not found (found: " . sprintf("%.2x %.2x", $b1, $b2) . ") - '$AICPUPM_file' can not be patched with spec. file '$spec_file'\n";
    } else {
        # wrmsr found - patch it
        # seek to our pos again
        seek($AICPUPM, $position, SEEK_SET) || die "ERROR: position $posHex can not be reached: $!\n";
        
        # and write 2 NOPs in there
        print $AICPUPM pack("C", $NOP), pack("C", $NOP);
        printf("%.2x%.2x wrmsr => %.2x%.2x nop nop - OK\n", $b1, $b2, $NOP, $NOP);
    }
    $num++;
}

# closing file
close($AICPUPM);

print "All done. Bye :)\n";
